# !/usr/bin/python
# coding=utf-8
import os
from typing import List, Tuple, Union, Dict, Any, Optional

try:
    import pymel.core as pm
except ImportError as error:
    print(__file__, error)
import pythontk as ptk

# from this package:
# from mayatk.core_utils import CoreUtils
# from mayatk.node_utils import NodeUtils
# from mayatk.env_utils import EnvUtils


class LightUtils(ptk.HelpMixin): ...


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    ...

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------
